/*********************************************
 * Incremental solution
 * Maximize number of patients served
 * Time window relaxation only 
 *********************************************/
 
 //Master problem.  Assign nurses to patients.

 // ***************************
 // User defined values
 // ***
 int nOldPatients = 54;      // Must be no greater than nPreAssignPatients
 int nNewPatients = 6;      // Must be no greater than number of patients
 int nAvailableNurses = 8;  // Nurses available for new patients
 // ***************************
 //
 
 int nPatients = ...;             // number of patient visits, including start & end home bases for nurses 
 int nNurses = ...;               // number of nurses
 int nDays = ...;                 // number of days to be scheduled
 
// It is assumed that the first nNurses visits are the start home bases of the nurses (in any order)
// and the last nNurses visits are the end home bases of the nurses (in any order).
// Home bases are matched with nurses in the array base[i][1..2].
 
 range Patients = 1..nPatients;
 range Nurses = 1..nNurses;
 range Days = 1..nDays;
 int release[Patients] = ...;     // patient visit release times 
 int deadline[Patients] = ...;    // patient visit deadlines
 int duration[Patients] = ...;    // patient visit durations (normally zero for home bases)
 int dayReq[Patients] = ...;        // number of days for visits 
 int code[Patients] = ...;        // constraint code
 // Codes:
 // 0: no constraints
 // 1: at least 1 day between visits
 // 2: at least 2 days between visits
 // 3: at leas 3 days between visits
 // Start and end home base of each nurse.  
 int base[Nurses][1..2] = ...;    // origin and final destination of each nurse    
 int homeBase[1..2];              // home bases passed to subproblem
 // Travel times must be given between patients and between each nurse home base (start and end) and patients.
 // Also between each nurse's start home and end home (zero if they are the same location).  
 // Times need not be symmetric.
 // Times between start home bases are not used, nor between end home bases. 
 int travelTime[Patients][Patients] = ...;
 int assignBar[Patients];         // patient assignments passed to subproblem
 // cost of assignning each nurse to each patient. 
 // float cost[Nurses][Patients] = ...;   
 // visit durations augmented by min time from or to previous or next visit (including end home base).
 int augmentedDurationFrom[Nurses][Patients][Days];
 int augmentedDurationTo[Nurses][Patients][Days];  
 int startBase[j in Patients] = 0;  // indicates whether a visit is a start base for some nurse.
 int endBase[j in Patients] = 0;    // indicates whether a visit is an end base for some nurse.
 // Benders cuts stored in this tuple.  Nogood contains coefficiencts.
 tuple bendersCut {
   	int nurse;
   	int day;
   	int rhs;
   	int nogood[Patients];
 }
 {bendersCut} BendersCuts = ...;
 
 int minTravelTime = 10000;
 // Data for multicommodity flow relaxation
 // Set of visits to patients only, excluing nurse home bases
 setof(int) PatientsExceptBase1 = { j | j in Patients : j>base[nNurses][1] };
 setof(int) PatientsExceptBase2 = { j | j in Patients : j<base[1][2] };
 setof(int) PatientsOnly = {j | j in PatientsExceptBase1 : j<base[1][2] };
 // Morning and evening patients.
 setof(int) MorningPatients = {j | j in PatientsOnly : deadline[j] <= 72 };
 setof(int) EveningPatients = {j | j in PatientsOnly : release[j] >= 72 };
 // Precomputed heuristic solution
 int preAssignNurse[PatientsOnly] = ...;
 int preAssignDays[PatientsOnly][Days] = ...;
 int nPreAssignPatients = 0;   // Number of patients assigned by heuristic algorithm.  Deduced from input file.
 //
 int assignNurseLowerBound[Nurses][PatientsOnly];
 int assignNurseUpperBound[Nurses][PatientsOnly];
 int assignLowerBound[Nurses][PatientsOnly][Days];
 int assignUpperBound[Nurses][PatientsOnly][Days];
 
 // Decision variables 
 // Whether nurse i is assigned to patient j
 dvar int assignNurse[Nurses][Patients] in 0..1;
 // Whether nurse i assigned to patient j on day d
 dvar int assign[Nurses][Patients][Days] in 0..1;
 // Whether patient is assigned a nurse
 dvar int assigned[Patients] in 0..1;
 
 execute {
   	// First compute nPreAssignPatients  
   	for(var j in PatientsOnly) {
     	if(preAssignNurse[j] > 0) nPreAssignPatients = nPreAssignPatients + 1;   
   	}
   	// Bound variables that are fixed by prescheduled patients, patients that are not part of the problem,
   	// or unavailable nurses. 
   	for(var i in Nurses) {
      	for(var j in PatientsOnly) {
        	if(j <= nOldPatients + nNurses) {
          	// These variables are fixed to 1 because they are prescheduled.        
          		if(preAssignNurse[j] == i) { 
            		assignNurseLowerBound[i][j] = 1;
            		assignNurseUpperBound[i][j] = 1;
            		for(var d in Days) {
              			assignLowerBound[i][j][d] = preAssignDays[j][d];
              			assignUpperBound[i][j][d] = 1;        
            		}
          		}
          		else {
            		assignNurseLowerBound[i][j] = 0;
            		assignNurseUpperBound[i][j] = 0;
            		for(var d in Days) {
              			assignLowerBound[i][j][d] = 0;
              			assignUpperBound[i][j][d] = 0;            
            		}          
          		}
        	}
        	else {
         		if(j <= nOldPatients+nNewPatients+nNurses) {
            		// These variables are free because they correspond to new patients for available nurses.
            		if(i <= nAvailableNurses) {
              			assignNurseLowerBound[i][j] = 0;
              			assignNurseUpperBound[i][j] = 1;
              			for(var d in Days) {
                			assignLowerBound[i][j][d] = 0;
                			assignUpperBound[i][j][d] = 1;
              			}                         
            		}
            		// These variables are fixed to zero for nurses unavailable for new patients.
            		else {
              			assignNurseLowerBound[i][j] = 0;
              			assignNurseUpperBound[i][j] = 0;
              			for(var d in Days) {
                			assignLowerBound[i][j][d] = 0;
                			assignUpperBound[i][j][d] = 0;
              			}               
            		}          
          		}
          		else {
            		// These variables are fixed to zero because the patients are not in the problem
            		assignNurseLowerBound[i][j] = 0;
            		assignNurseUpperBound[i][j] = 0;
            		for(var d in Days) {
              			assignLowerBound[i][j][d] = 0;
              			assignUpperBound[i][j][d] = 0;            
            		}          
          		}
        	} 
    	}
   	}    
   
   	// Identify visits that are a start base or end base for some nurse, and set day requirement to number of days.
   	for(var i in Nurses) {
     	startBase[base[i][1]] = 1;
     	dayReq[base[i][1]] = nDays;
     	endBase[base[i][2]] = 1;   
     	dayReq[base[i][2]] = nDays;  
   	} 
   	// Augment durations to include min travel time to or from next or previous visit.
   	for(var i in Nurses) {
   		for(var d in Days) {
   			for(var j in PatientsOnly) {
   				if (assignUpperBound[i][j][d] == 1) {   			
   					// Travel time from previous.
   					minTravelTime = 10000;	
   					// Start base first.
   					if (travelTime[base[i][1]][j] < minTravelTime) minTravelTime = travelTime[base[i][1]][j];
   					// Now other patients.
   					for (var jj in PatientsOnly) {
   						if (j != jj) {
         					if (assignUpperBound[i][jj][d] == 1) {         				
           						if (travelTime[jj][j] < minTravelTime) minTravelTime = travelTime[jj][j];
              				}           					   					
   						}
     				}
     				augmentedDurationFrom[i][j][d] = duration[j] + minTravelTime;
     				// Travel time to next.
     				minTravelTime = 10000;	
   					// End base 
   					if (travelTime[j][base[i][2]] < minTravelTime) minTravelTime = travelTime[base[i][1]][j];
   					// Now other patients
   					for (jj in PatientsOnly) {
   						if (j != jj) {
         					if (assignUpperBound[i][jj][d] == 1) {         				
           						if (travelTime[j][jj] < minTravelTime) minTravelTime = travelTime[j][jj];
              				}           					   					
   						}
     				}
     				augmentedDurationTo[i][j][d] = duration[j] + minTravelTime;
       			}     				   				   			
   			}
   		}          	     		     
   	//}
   	//for(var j in Patients) {
   		//writeln("Patient ",j);   	
   		//for(var i in Nurses) {
   			//write("Nurse ",i," ");
   			//for(var d in Days) write(augmentedDuration[i][j][d]," ");
   			//writeln();
   		//}   		
   	}   	
 }
 
 maximize
   	// Maximize number of patients served.
   	sum(j in PatientsOnly) assigned[j];
 
 subject to {
 
   	forall(j in Patients) {
     	// Patient assigned a nurse. 
     	sum(i in Nurses) assignNurse[i][j] == assigned[j];
     	// Number of days per patient 
     	sum(i in Nurses, d in Days) assign[i][j][d] == dayReq[j] * assigned[j];
     	// Same nurse assign to a given patient on all days.
     	forall(i in Nurses) 
       		forall (d in Days) 
         		assign[i][j][d] <= assignNurse[i][j];           
   	}     
   	// Nurses must be assigned their start and end home bases on each day.
   	forall(i in Nurses) {
     	forall(d in Days) {   
       		assign[i][base[i][1]][d] == 1;
       		assign[i][base[i][2]][d] == 1;
     	}     
   	}   
   	// Code is minimum separation.
   	// Logic for 7-day week.
   	//forall(j in PatientsOnly) {
    	//forall(i in Nurses) {     
       		//forall(d in Days) {
         		//forall(sep in 1..code[j]) {
           			//assign[i][j][((d+sep-1) mod nDays) + 1] <= 1 - assign[i][j][d];
         		//}           
       		//}         
     	//}     
   	//}
   	// Logic for 5-day week.
   	forall(j in PatientsOnly) {
     	forall(i in Nurses) {     
       		forall(d in Days : d!=nDays) {
         		forall(sep in 1..code[j] : d+sep<=nDays) {
           			assign[i][j][d+sep] <= 1 - assign[i][j][d];
         		}           
       		}         
     	}     
   	}
   
   	//Benders cuts.
   	forall(c in BendersCuts) {
     	sum(j in Patients) c.nogood[j]*(1 - assign[c.nurse][j][c.day]) >= c.rhs;
     	// Additional cuts
     	sum(j in Patients, d in Days:d!=c.day) c.nogood[j]*(1 - assign[c.nurse][j][d]) >= nDays - 1;
     	//forall(d in Days : d != c.day) {
     		//sum(j in Patients) c.nogood[j]*(1 - assign[c.nurse][j][d]) >= c.rhs;
     	//}     
   	} 
   
   	// Prescheduled and nonparticipating patients.
   	forall(i in Nurses) {
     	forall(j in PatientsOnly) {
       		assignNurse[i][j] >= assignNurseLowerBound[i][j];
       		assignNurse[i][j] <= assignNurseUpperBound[i][j];
       		forall(d in Days) {
         		assign[i][j][d] >= assignLowerBound[i][j][d];
         		assign[i][j][d] <= assignUpperBound[i][j][d];
       		}
     	}        
   	}
   
   	//  Time window relaxation.  Restore lines with *
   	forall(i in Nurses) {
     	forall(d in Days) {
     	 	// Morning and evening patients
       		//augmentedDuration[i][base[i][1]][d]*assign[i][base[i][1]][d] + sum(j in MorningPatients) augmentedDuration[i][j][d]*assign[i][j][d] <= 72 - release[base[i][1]];
       		sum(j in MorningPatients) augmentedDurationFrom[i][j][d]*assign[i][j][d] <= 72 - release[base[i][1]];
       		//augmentedDuration[i][base[i][1]][d]*assign[i][base[i][1]][d] <= 72 - release[base[i][1]];
       		//augmentedDuration[i][base[i][1]][d]*assign[i][base[i][1]][d] + sum(j in EveningPatients) augmentedDuration[i][j][d]*assign[i][j][d] <= deadline[base[i][2]] - 72;
       		sum(j in EveningPatients) augmentedDurationTo[i][j][d]*assign[i][j][d] <= deadline[base[i][2]] - 72;
       		//augmentedDuration[i][base[i][1]][d]*assign[i][base[i][1]][d] <= deadline[base[i][2]] - 72;
       		// Individual patient time windows.
       		forall(j in PatientsOnly) {
       			(release[base[i][1]] + augmentedDurationFrom[i][j][d])*assign[i][j][d] <= deadline[j];       		
         		(release[j] + augmentedDurationTo[i][j][d])*assign[i][j][d] <= deadline[base[i][2]];
       		}       		
     	}     
   	}
 } 
 
 
 main {
   	thisOplModel.generate();
   
   	var masterDef = thisOplModel.modelDefinition;
   	var masterCplex = cplex;
   	var masterData = thisOplModel.dataElements; 
   
   	var subSource = new IloOplModelSource("homeCare6Sub.mod");
   	var subDef = new IloOplModelDefinition(subSource);
   	var subData = new IloOplDataElements();
   	var subCP = new IloCP();
   	var infeasible = 1;
   	var whichNurse = 1;
   	var iter = 1;
   	var minTravelTime;
   	var solved = 1;
   	var longPrint = 0;
     
   	// Check values set by user.
   	var masterOpl = new IloOplModel(masterDef,masterCplex);
   	masterOpl.addDataSource(masterData);
   	masterOpl.generate();
   	if (masterOpl.nOldPatients <= masterOpl.nPreAssignPatients) {
     	writeln(masterOpl.nOldPatients," old patients selected from ",masterOpl.nPreAssignPatients," prescheduled patients.");   
   	}
   	else {
     	writeln("WARNING: ",masterOpl.nOldPatients," old patients selected exceeds ",masterOpl.nPreAssignPatients," prescheduled patients.");   
   	}
   	if (masterOpl.nNewPatients <= masterOpl.nPatients - 2*masterOpl.nNurses - masterOpl.nOldPatients) {
     	writeln(masterOpl.nNewPatients," new patients selected from ",masterOpl.nPatients - 2*masterOpl.nNurses - masterOpl.nOldPatients," available.");   
   	}
   	else {
     	writeln("ERROR: ",masterOpl.nNewPatients, " new patients exceeds ",masterOpl.nPatients - 2*masterOpl.nNurses - masterOpl.nOldPatients," available.");   
   	}
   	if (masterOpl.nAvailableNurses <= masterOpl.nNurses) {
     	writeln(masterOpl.nAvailableNurses," of ",masterOpl.nNurses," nurses are available for new patients.");   
   	}
   	else {
     	writeln("ERROR: ",masterOpl.nAvailableNurses," nurses selected for new patients, only ",masterOpl.nNurses," available.");
   	}
   	masterOpl.end();
 
   	// Iterate until all nurse subproblems are feasible.
   	while (infeasible) {
     	infeasible = 0;
     	var masterOpl = new IloOplModel(masterDef,masterCplex);
     	masterOpl.addDataSource(masterData);
     	masterOpl.generate();
     	
     	// Solve the master problem.
     	//writeln(masterOpl.assignNurseUpperBound);
     	if(longPrint==1) writeln(iter,".  Solve master");
     	if (masterCplex.solve()) {
     		//for(var i in masterOpl.Nurses) {
     			//writeln("Nurse ",i);     		
     			//for(var d in masterOpl.Days) {
     				//write(" Day ",d,":");     			
     				//for(var j in masterOpl.Patients) {
     					//write(" ",masterOpl.augmentedDurationFrom[i][j][d],",",masterOpl.augmentedDurationTo[i][j][d]);     				
     				//}
     				//writeln();     			
     			//}	
     		//}     	
     	
       		if (longPrint==1) {
         		writeln("Objective: ",masterCplex.getObjValue(),".  Assignments:");   
         		//writeln("       ",masterOpl.assign);
         		for(var d in masterOpl.Days) {
           			writeln("Day ",d,":");
           			for(var i in masterOpl.Nurses) {
             			write(" Nurse ",i,": ");
             				for(var j in masterOpl.Patients) {
               					write(masterOpl.assign[i][j][d]);           
             				}         
             				write(" Patients: ");
             				for(var j in masterOpl.PatientsOnly) {
               					if(masterOpl.assign[i][j][d]==1) write(j," ");           
             				}
             				writeln();
           				}           
         			}       
       			}
       			else {
         			write(iter,".  Value: ",masterCplex.getObjValue()," Infeasible: ");       
       			}       
     		}   
     		else {
       			// If master is infeasible, stop.     
       			writeln("No solution");
       			solved = 0;
       			masterOpl.end();
       			break;   
     		}
     
     		// Loop through the days.
     		for(d in masterOpl.Days) {
       			whichNurse = 1;
       			// Loop through the nurses.
       			while (whichNurse <= masterOpl.nNurses) {
         			subData.nPatients = masterOpl.nPatients;
         			subData.nNurses = masterOpl.nNurses;
         			subData.release = masterOpl.release;
         			subData.deadline = masterOpl.deadline;
         			subData.duration = masterOpl.duration;
         			//subData.cost = masterOpl.cost;
         			subData.homeBase = masterOpl.homeBase;
         			for(var k=1; k<=2; k++) {
           				masterOpl.homeBase[k] = masterOpl.base[whichNurse][k];    
           				subData.homeBase[k] = masterOpl.homeBase[k];
         			}       
         			subData.travelTime = masterOpl.travelTime;
         			subData.assignBar = masterOpl.assignBar;
         			subData.release = masterOpl.release;
         			subData.deadline = masterOpl.deadline;
         			for(var j in masterOpl.Patients) {
           				// Tell the subproblem which patients are assigned to this nurse.            
           				masterOpl.assignBar[j] = masterOpl.assign[whichNurse][j][d];       
           				subData.assignBar[j] = masterOpl.assignBar[j];            
         			}
       
         			var subOpl = new IloOplModel(subDef, subCP);
         			subOpl.addDataSource(subData);
         			subOpl.generate();
   
         			//writeln();
         			if(longPrint==1) write("Day ",d,", Nurse ",whichNurse," subproblem");
         			if(subCP.solve()) {
           				// Subproblem is feasible.       
           				if(longPrint==1) writeln(" feasible");     
           				//writeln(d,", ",whichNurse," ",subData.assignBar);
           				//writeln(subOpl.s);
           				//writeln(subOpl.start.solutionValue);
           				subOpl.end(); 
         			}
         			else {
           				// Subproblem is infeasible.  Must generate a Benders cut.        
           				if(longPrint==1) write(" infeasible.  Visits deleted: ");
           				else write(d,":",whichNurse," ");   
           				infeasible = 1;
           				//writeln("thesePatients = ",subOpl.thesePatients);
           				subOpl.end();
           				// Try to strengthen the cut by removing patients one at a time.
           				for(var j in masterOpl.PatientsOnly) {         
             				if(masterOpl.assignBar[j]==1) {
               					if(masterOpl.assignLowerBound[whichNurse][j][d] == 0) {       
                 					if(1 == 1) {    
                   						masterOpl.assignBar[j] = 0;
                   						subData = new IloOplDataElements();
                   						subCP = new IloCP();        
                   						subData.nPatients = masterOpl.nPatients;
                   						subData.nNurses = masterOpl.nNurses;
                   						subData.release = masterOpl.release;
                   						subData.deadline = masterOpl.deadline;
                   						subData.duration = masterOpl.duration;
                   						//subData.cost = masterOpl.cost;
                   						subData.homeBase = masterOpl.homeBase;
                   						for(var jj=1; jj<=2; jj++) {
                     						masterOpl.homeBase[jj] = masterOpl.base[whichNurse][jj];    
                     						subData.homeBase[jj] = masterOpl.homeBase[jj];
                   						}                 
                   						subData.travelTime = masterOpl.travelTime;
                   						subData.assignBar = masterOpl.assignBar;
                   						//subData.whichNurse = whichNurse;
                   						for(var jj in masterOpl.Patients) {
                     						// Tell the subproblem which patients are assigned to this nurse.            
                     						subData.assignBar[jj] = masterOpl.assignBar[jj];            
                   						}
                   						//subData.assignBar[j] = 0;    
                   						var subOpl = new IloOplModel(subDef, subCP);
                   						subOpl.addDataSource(subData);
                   						subOpl.generate();
                   						//write(subOpl.assignBar);
                   						if(subCP.solve()) {
                     						masterOpl.assignBar[j] = 1;
                     						//writeln("  feasible");        
                   						}      
                   						else {
                     						//writeln("  infeasible");      
                     						masterOpl.assignBar[j] = 0;
                     						if(longPrint==1) write(j," ");      
                   						}
                   						//write(subOpl.thesePatients);
                   						subOpl.end(); 
                 					}
               					}
             				}
           				}
           				if(longPrint==1) writeln();                  
           				//writeln(subOpl.Patients);
           				masterData.BendersCuts.add(whichNurse,d,1,masterOpl.assignBar);
           				//writeln(whichNurse," ",subData.assignBar);
         			}
         			whichNurse = whichNurse + 1;
   					//    subSource = new IloOplModelSource("homeCareSub.mod");
   					//    subDef = new IloOplModelDefinition(subSource);
         			subData = new IloOplDataElements();
         			subCP = new IloCP();
       			}
     		}          
     		if(infeasible==1) masterOpl.end();     		
     		iter = iter + 1;
     		writeln();
   		}    
   		if(solved==1) {
   			writeln("Problem solved.");
   			var outfile = new IloOplOutputFile("patientAssignments.txt");
   			// Write nurse assigned to each patient. 
   			for(var j in masterOpl.PatientsOnly) {
    			for(var i in masterOpl.Nurses) {
    				if(masterOpl.assignNurse[i][j] == 1) outfile.write(i," ");    		
    			}
    		}
    		outfile.writeln();
    		// Write days to which patients are assigned (0 or 1 for each day) 
    		for(var j in masterOpl.PatientsOnly) {
    			for(d in masterOpl.Days) {
    				for(var i in masterOpl.Nurses) {    	
    					if(masterOpl.assignNurse[i][j] == 1) outfile.write(masterOpl.assign[i][j][d]," ");
    				}    		   	
    			}
    			outfile.writeln();
   			}
   			outfile.close();
   			masterOpl.end();    	
    	}
 	}